type var = (Strng.t * int);;
type t = Var of var | Abs of (var * t) | App of (t * t);;

let var c =  Var ([c], 0);;
let abs xs body = Lst.fold (fun x xs -> Abs (([x], 0), xs)) body xs;;
let app ts = Lst.fold_left1 (fun s t -> App (s, t)) ts;;
let rec max_index = function
 | Var (_, i) -> i
 | Abs ((_, i), u) -> max i (max_index u)
 | App (u, v) -> max (max_index u) (max_index v)
;;
let rename_bound i t =
 let inc (id, i) j = (id, i + j) in
 let rec rename_bound i bvs = function
  | Var x as v -> if St.mem x bvs then Var (inc x i) else v
  | Abs (x, u) -> Abs (inc x i, rename_bound i (St.insert x bvs) u)
  | App (u, v) -> App (rename_bound i bvs u, rename_bound i bvs v)
 in
 rename_bound i St.empty t
;;
let rec substitute x t = function
 | Var y as v -> if y = x then t else v
 | Abs (y, u) as s -> if y = x then s else Abs (y, substitute x t u)
 | App (u, v) -> App (substitute x t u, substitute x t v)
;;
let beta = function
 | App (Abs (x, u), v) -> substitute x v (rename_bound (max_index v + 1) u)
 | _ -> failwith "Lambda.beta: not a redex"
;;
let is_redex = function
 | App (Abs _, _) -> true
 | _ -> false
;;
let outer stop t =
 let rec outer t = if stop t then (false, t) else match t with
  | Var _ as x -> (false, x)
  | Abs (x, u) -> let (r, u) = outer u in (r, Abs (x, u))
  | App (u, v) as t ->
   if is_redex t then
    (true, beta t)
   else
    let (r, u) = outer u in
    if r then (r, App (u, v)) else let (r, v) = outer v in (r, App (u, v))
 in snd (outer t)
;;
let inner stop t =
 let rec inner t = if stop t then (false, t) else match t with
  | Var _ as x -> (false, x)
  | Abs (x, u) -> let (r, u) = inner u in (r, Abs (x, u))
  | App (u, v) ->
   let (r, u) = inner u in
   if r then (r, App (u, v)) else let (r, v) = inner v in
   if r then (r, App (u, v)) else let t = App (u, v) in
   if is_redex t then (true, beta t) else (false, t)
 in snd (inner t)
;;
let rec vars = function
 | Var x -> St.singleton x
 | Abs (x, u) -> St.insert x (vars u)
 | App (u, v) -> St.union (vars u) (vars v)
;;
let rec fvars = function
 | Var x -> St.singleton x
 | Abs (x, u) -> St.diff (fvars u) (St.singleton x)
 | App (u, v) -> St.union (fvars u) (fvars v)
;;
let rec bvars t = St.diff (vars t) (fvars t);;
let is_app = function
 | App _ -> true
 | _ -> false
;;
let rec is_whnf = function
 | App _ -> false
 | _ -> true
;;
let rec is_nf = function
 | Var _ -> true
 | Abs (x, u) -> is_nf u
 | App (u, v) as t -> if is_redex t then false else is_nf u && is_nf v
;;
let is_closed t = St.is_empty (fvars t);;
let var2strng (id, i) =
 if i < 0 then id else Lst.append id (Strng.of_int i)
;;
let rec to_strng = function
 | Var x -> var2strng x
 | Abs (x, u) ->
  '(' :: '%' :: Lst.append (Lst.append (var2strng x) ('.' :: to_strng u)) [')']
 | App (u, v) ->
  '(' :: Lst.append (Lst.append (to_strng u) (' ' :: to_strng v)) [')']
;;
let rec sub = function
 | Var _ as x -> St.singleton x
 | Abs (x, u) as t -> St.insert t (sub u)
 | App (u, v) as t -> St.insert t (St.union (sub u) (sub v))
;;
let rec sub_list = function
 | Var _ as x -> [x]
 | Abs (x, u) as t -> t :: sub_list u
 | App (u, v) as t -> t :: Lst.append (sub_list u) (sub_list v)
;;
let redexes t = St.filter (is_redex) (sub t);;
let redexes_list t = Lst.filter (is_redex) (sub_list t);;
let by_name t =
 let u = Lst.until is_whnf (outer is_whnf) t in
 Lst.until is_nf (outer is_nf) u
;;
let by_value t =
 let u = Lst.until is_whnf (inner is_whnf) t in
 Lst.until is_nf (inner is_nf) u
;;
let print fmt t =
 let pvar fmt (id, i) =
  let rec pid fmt = function
   | [] -> ()
   | c :: cs -> Format.fprintf fmt "%c%a" c pid cs
  in
  Format.fprintf fmt "%a%i" pid id i
 in
 let rec pabs fmt = function
  | Abs (x, (Abs _ as u)) -> Format.fprintf fmt "%a%a" pvar x pabs u
  | Abs (x, u) -> Format.fprintf fmt "%a.%a" pvar x print u
  | t -> print fmt t
 and print fmt = function
  | Var x -> pvar fmt x
  | Abs _ as t -> Format.fprintf fmt "%%%a" pabs t
  | App (Var _ as u, (Var _ as v))
  | App (App _ as u, (Var _ as v)) ->
   Format.fprintf fmt "%a %a" print u print v
  | App (Var _ as u, (Abs _ as v))
  | App (Var _ as u, (App _ as v))
  | App (App _ as u, (Abs _ as v))
  | App (App _ as u, (App _ as v)) ->
   Format.fprintf fmt "%a (%a)" print u print v
  | App (Abs _ as u, (Var _ as v)) ->
   Format.fprintf fmt "(%a) %a" print u print v
  | App (Abs _ as u, (Abs _ as v))
  | App (Abs _ as u, (App _ as v)) ->
   Format.fprintf fmt "(%a) (%a)" print u print v
 in
 print fmt t
;;
let toplevel_printer = print Format.std_formatter;;
